package ru.dbotthepony.kstarbound.io.json

import com.google.gson.TypeAdapter
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.util2d.AABBi
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kvector.vector.nint.Vector2i

object AABBTypeAdapter : TypeAdapter<AABB>() {
	override fun write(out: JsonWriter, value: AABB) {
		`out`.beginArray()
		Vector2dTypeAdapter.write(out, value.mins)
		Vector2dTypeAdapter.write(out, value.maxs)
		`out`.endArray()
	}

	override fun read(`in`: JsonReader): AABB {
		val (x1, x2) = Vector2dTypeAdapter.read(`in`)
		val (y1, y2) = Vector2dTypeAdapter.read(`in`)

		val xMins = x1.coerceAtMost(x2)
		val xMaxs = x1.coerceAtLeast(x2)

		val yMins = y1.coerceAtMost(y2)
		val yMaxs = y1.coerceAtLeast(y2)

		return AABB(
			Vector2d(xMins, yMins),
			Vector2d(xMaxs, yMaxs),
		)
	}
}

object AABBiTypeAdapter : TypeAdapter<AABBi>() {
	override fun write(out: JsonWriter, value: AABBi) {
		`out`.beginArray()
		Vector2iTypeAdapter.write(out, value.mins)
		Vector2iTypeAdapter.write(out, value.maxs)
		`out`.endArray()
	}

	override fun read(`in`: JsonReader): AABBi {
		val (x1, x2) = Vector2iTypeAdapter.read(`in`)
		val (y1, y2) = Vector2iTypeAdapter.read(`in`)

		val xMins = x1.coerceAtMost(x2)
		val xMaxs = x1.coerceAtLeast(x2)

		val yMins = y1.coerceAtMost(y2)
		val yMaxs = y1.coerceAtLeast(y2)

		return AABBi(
			Vector2i(xMins, yMins),
			Vector2i(xMaxs, yMaxs),
		)
	}
}

