package ru.dbotthepony.kstarbound.io.json

import com.google.common.collect.ImmutableList
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kstarbound.Starbound

fun <T : Any> TypeAdapter<T>.asList(): TypeAdapter<List<T>> {
	return ListAdapter(this)
}

class ListAdapter<T : Any>(val elementAdapter: TypeAdapter<T>, val valueTransformer: (T) -> T = { it }) : TypeAdapter<List<T>>() {
	constructor(type: Class<T>) : this(LazyTypeProvider(type))

	override fun write(out: JsonWriter, value: List<T>) {
		out.beginArray()

		for (v in value) {
			elementAdapter.write(out, v)
		}

		out.endArray()
	}

	override fun read(reader: JsonReader): List<T> {
		reader.beginArray()

		val builder = ImmutableList.builder<T>()

		while (reader.peek() != JsonToken.END_ARRAY) {
			val readObject = elementAdapter.read(reader) ?: throw JsonSyntaxException("List does not accept nulls")
			builder.add(valueTransformer(readObject))
		}

		reader.endArray()

		return builder.build()
	}
}
